# cronjob.tcl --
#    Script to run a series of periodic commands
#
#    General information:
#    This script gives some of the functionality of the cron command on
#    UNIX. It will read a source file at start-up containing the
#    commands to run periodically.
#
#    Note:
#    These commands can only be Tcl procedures. If they have arguments,
#    then these should be fixed and given in the same line.
#
#    In the source file the command Job adds a scheduled procedure
#    to the list. It takes the following arguments:
#    start-time period command
#
# Author: Arjen Markus
#

# Cronjob --
#    Namespace for the commands
#
namespace eval ::Cronjob {
   variable nojobs     0
   variable jobs
   variable logfile
   variable alivefile
}

# Log --
#    Write a log message to the log file
#
# Arguments:
#    message     Message text to be written
#
# Return value:
#    None
#
# Side effects:
#    Write the text to the log file and flush the file
#
proc ::Cronjob::Log { message } {
   variable logfile

   set timestring [clock format [clock seconds] -format "%Y/%m/%d %H:%M"]
   puts $logfile "$timestring $message"
   flush $logfile
}

# Job --
#    Schedule a job for periodic execution
#
# Arguments:
#    start_time  Time of the day to start the first job
#                (must be in the format hh:mm)
#    period      Period to repeat the job (in hours and minutes, hh:mm)
#    args        Command to perform and any (fixed) arguments
#
# Return value:
#    None
#
# Side effects:
#    Add the job to the list of jobs
#
proc ::Cronjob::Job { start_time period command } {
   variable jobs
   variable nojobs

   regexp {([0-9]+):([0-9]+)} $period match hours minutes
   set nosecs [ expr $hours*3600+$minutes*60 ]

   set jobs($nojobs,start_time) $start_time
   set jobs($nojobs,period)     $nosecs
   set jobs($nojobs,command)    $command

   set next_time [ clock scan "today $start_time" ]
   set curr_time [ clock seconds ]

   while { $next_time < $curr_time } {
      incr next_time $jobs($nojobs,period)
   }

   set jobs($nojobs,next_time) $next_time

   incr nojobs

   return
}

# MainLoop --
#    Execute the scheduled jobs.
#
#    In a loop:
#    - Determine the next time a job has to start.
#    - Then calculate the time that is left until the
#      start. Sleep that long and execute the job.
#    - Repeat this indefinitely.
#
# Arguments:
#    None
#
# Return value:
#    None
#
# Side effects:
#    Whatever the commands do that are being executed per job
#
proc ::Cronjob::MainLoop { } {
   variable jobs
   variable nojobs
   variable next_job

   #
   # As there is bound to be a job tomorrow, take that as the
   # limit
   #
   OpenAuxFiles
   Log "Cronjob started"

   while { 1 } {
      set next_job  -1
      set next_time [clock scan "tomorrow" ]

      for { set i 0 } { $i < $nojobs } { incr i } {
         if { $jobs($i,next_time) < $next_time } {
            set next_time $jobs($i,next_time)
            set next_job  $i
         }
      }

      if { $next_job != -1 } {
         set sleep [ expr 1000 * ( $next_time - [ clock seconds ] ) ]
         if { $sleep > 0 } {
            puts "Sleep for [expr $sleep/1000] seconds..."
            after $sleep
         }

         incr jobs($next_job,next_time) $jobs($next_job,period)
         if { [ catch {
                   eval $jobs($next_job,command)
                } result ] != 0 } {
            puts stderr "Error occurred in job $next_job:\n $result"
            Log  "Error occurred in job $next_job:\n $result\nCode:\n\
$jobs($next_job,command)"
         }
      } else {
         puts stderr "Error: no job found!"
         break
      }
   }

   return
}

# OpenAuxFiles --
#    Open the auxiliary files
#
# Arguments:
#    None
#
# Return value:
#    None
#
# Side effects:
#    Opens two files for logging
#
proc ::Cronjob::OpenAuxFiles { } {
   variable logfile
   variable alivefile

   set logfile   [open "cronjob.log"   "w"]
   set alivefile [open "cronjob.alive" "w"]
}

# exit --
#    Redefined exit to make sure that the event is logged
#
# Arguments:
#    retvalue       Optional return value
#
# Return value:
#    None
#
# Side effects:
#    Writes a message to the log file, stops the process
#
rename exit _exit

proc ::Cronjob::exit { {retvalue 0} } {
   variable next_job

   Log "Exit from Cronjob in job $next_job"

   _exit $retvalue
}

namespace eval ::Cronjob {
   namespace export Log Job MainLoop exit
}

# main code --
#    Source the crontab file and start the main loop
#

# Open the log file, the "alive" file and set up two
# administrative jobs - one to write an "alive" message,
# the other to restart this file.
#
namespace eval ::Cronjob {

   namespace export Log Job MainLoop exit
}

namespace import ::Cronjob::*

Job 0:00 0:10 {
   set timestring [clock format [clock seconds] -format "%Y/%m/%d %H:%M"]
   puts $::Cronjob::alivefile "$timestring Cronjob alive"
   flush $::Cronjob::alivefile
}

Job 0:00 24:00 {
   close $::Cronjob::alivefile
   set ::Cronjob::alivefile [open "cronjob.alive" "w"]
}

#
# Start the actual cronjob
#
if { [ file exists "crontab.tcl" ] } {
   source crontab.tcl
}

MainLoop

Log "Normal exit of Cronjob"
_exit
